"""Provides functions to serialize materials."""
import json
import pathlib
from typing import Dict, Union

from ansys.materials.manager._models import Constant
from ansys.materials.manager.material import Material


def _material_as_dict(material: Material) -> Dict:
    d = {"name": material.name}
    if material.material_id is not None:
        d.update({"id": material.material_id})
    for model in material.models:
        if isinstance(model, Constant):
            propData = {"option": "constant", "value": model.value}
        else:
            propData = {"option": "ideal_gas"}
        d.update({model.name: propData})
    return d


def serialize_material(material: Material) -> str:
    """
    Output the JSON representation of a material in Fluent format to a string.

    Parameters
    ----------
    material : Material
        Material to serialize.

    Returns
    -------
    str
        String representation of the material in Fluent format.
    """
    d = _material_as_dict(material)
    return json.dumps(d)


def serialize_material_to_file(material: Material, file_name: Union[str, pathlib.Path]):
    """
    Output the JSON representation of a material in Fluent format to a file.

    Parameters
    ----------
    material : Material
        Material to serialize.
    file_name : Union[str, pathlib.Path]
        Name of the file to create.
    """
    sm = serialize_material(material)
    with open(file_name, "w", encoding="utf8") as f:
        f.write(sm)



from typing import Any, List, Tuple

from ansys.materials.manager._models import ModelValidationException, _BaseModel
from ansys.materials.manager._models._common._base import _FluentCore
from ansys.materials.manager._models._common._packages import SupportedPackage
from ansys.materials.manager._models._fluent.simple_properties import (
    property_codes as fluent_property_codes,
)
from ansys.materials.manager.material import Material


class IdealGas(_BaseModel):
    r"""
    Provides for creating an ideal gas model for fluid properties.

    This model can be applied to density and specific heat capacity properties within Fluent.
    It requires that the molar mass be set as a property and models the following equation:

    .. math::

     P * V = \frac{m * R * T}{M}
    """

    applicable_packages = SupportedPackage.FLUENT
    _name: str

    def __init__(self, name: str):
        """Create an ideal gas model for the Fluent solver."""
        self._name = name

    @property
    def name(self) -> str:
        """Name of the property modeled."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    def write_model(self, material: "Material", pyansys_session: Any) -> None:
        """
        Write this model to Fluent.

        This method should make some effort to validate the model state before writing.

        Parameters
        ----------
        material: Material
            Material object to associate with this model.
        pyansys_session: Any
            Configured instance of the PyAnsys session.
        """
        if not isinstance(pyansys_session, _FluentCore):
            raise TypeError(
                "This model is only supported by MAPDL and Fluent. Ensure that you have the correct"
                "type of the PyAnsys session."
            )

        is_ok, issues = self.validate_model()

        molar_mass_prop = material.get_model_by_name("molar mass")
        if len(molar_mass_prop) == 0:
            is_ok = False
            issues.append("Molar mass must be provided for the ideal gas model.")
        if not is_ok:
            raise ModelValidationException("\n".join(issues))

        fluent_property_code = fluent_property_codes[self._name.lower()]
        pyansys_session.setup.materials.fluid[material.name] = {
            fluent_property_code: {"option": "ideal-gas"}
        }

    def validate_model(self) -> "Tuple[bool, List[str]]":
        """
        Perform pre-flight validation of the model setup.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        failures = []
        is_ok = True

        if self._name is None or self._name == "":
            failures.append("Invalid property name")
            is_ok = False

        return is_ok, failures



"""Provides the ``MaterialDataParser`` class."""
from typing import Dict, List, Optional

from ansys.materials.manager._models import Constant, PiecewiseLinear, _BaseModel
from ansys.materials.manager._models._mapdl.simple_properties import property_codes
from ansys.materials.manager.util.common import FLOAT_VALUE_REGEX, MP_MATERIAL_HEADER_REGEX, np


class _MaterialDataParser:
    @staticmethod
    def parse_material(data: str, id_: int) -> List[_BaseModel]:
        """
        Parse the response from an `MPLIST` command.

        Creates a list of material models applied to the material.

        Parameters
        ----------
        data : str
            String response from the `MPLIST` command.
        id_ : int
            Material identity to be parsed.

        Returns
        -------
        List[_BaseModel]
            List of models assigned to this material.
        """
        data_section = _MaterialDataParser._get_mp_section_with_id(data, id_)
        return _MaterialDataParser._process_material(data_section)

    @staticmethod
    def _get_mp_section_with_id(data: str, id_: int) -> List[str]:
        """
        Extract material property information for the specified material identity.

        Parameters
        ----------
        data : str
            String response from the `MPLIST` command.
        id_ : int
            Material identity to be extracted.

        Returns
        -------
        List[str]
            Relevant section of the data input, split on newlines.
        """
        material_ids = map(int, MP_MATERIAL_HEADER_REGEX.findall(data))
        if id_ not in material_ids:
            raise IndexError(f"Material with ID {id_} not found in data")

        relevant_lines = []
        reading_correct_material = False
        for line in data.splitlines():
            stripped_line = line.strip()
            if stripped_line.startswith("MATERIAL"):
                match = MP_MATERIAL_HEADER_REGEX.match(stripped_line)
                if match:
                    current_id = int(match.groups()[0])
                    if current_id == id_:
                        reading_correct_material = True
                    else:
                        reading_correct_material = False
            elif reading_correct_material:
                relevant_lines.append(line)

        return relevant_lines

    @staticmethod
    def _process_material(material_data: List[str]) -> List[_BaseModel]:
        """
        Deserialize a material into a list of material models.

        Parameters
        ----------
        material_data : List[str]
            List of lines containing property data for one material.

        Returns
        -------
        List[_BaseModel]
            List of models applied to this material.
        """
        property_data: List[_BaseModel] = []
        property_lines: Dict[str, List[str]] = {}
        reference_temperature = None
        current_property_code = None
        lines = [line.strip() for line in material_data if line.strip()]
        for line in lines:
            if line.startswith("TEMP"):
                current_property_code = _MaterialDataParser._process_property_header(line)
                property_lines[current_property_code] = []
            elif line.startswith("REFT"):
                temp_string = line.split("=")[1]
                temp_val_match = FLOAT_VALUE_REGEX.search(temp_string)
                assert temp_val_match is not None, "Invalid material input"
                reference_temperature = float(temp_val_match.group(0))
            else:
                assert current_property_code is not None, "Invalid material input"
                property_lines[current_property_code].append(line)
        for name, value in property_lines.items():
            property_data.append(_MaterialDataParser._process_property(name, value))
        if reference_temperature is not None:
            property_data.append(Constant("Strain Reference Temperature", reference_temperature))
        return property_data

    @staticmethod
    def _process_property_header(header_line: str) -> str:
        """
        Deserialize a property header line into the relevant named property.

        Parameters
        ----------
        header_line : str
            Material property header line.

        Returns
        -------
        str
            Corresponding standard name for the property.

        Raises
        ------
        KeyError
            If the header line specifies an unknown property.
        IndexError
            If the header line does not match the expected format.
        """
        stripped_header_line = header_line.strip()
        try:
            property_name = stripped_header_line[4:].strip().split(" ")[0]
        except IndexError:
            raise IndexError("Invalid property header line")
        try:
            return next(k for k, v in property_codes.items() if v == property_name.upper())
        except StopIteration:
            raise KeyError(f"Invalid property: '{property_name}'")

    @staticmethod
    def _process_property(property_name: str, property_data: List[str]) -> _BaseModel:
        """
        Deserialize the property data into a model representing the property.

        Single values are deserialized into Constant models, arrays are deserialized
        into PiecewiseLinear models.

        Parameters
        ----------
        property_name : str
            Name of the property to be deserialized.
        property_data : List[str]
            Property data section to be deserialized.

        Returns
        -------
        _BaseModel
            Deserialized model, either a Constant or PiecewiseLinear model at this point.
        """
        model: Optional[_BaseModel] = None
        if len(property_data) == 1:
            match = FLOAT_VALUE_REGEX.search(property_data[0])
            if match:
                property_value = float(match.group(0))
                model = Constant(property_name, property_value)
        else:
            property_value = []
            parameter_value = []
            for data_line in property_data:
                line_values = FLOAT_VALUE_REGEX.findall(data_line)
                parameter_value.append(float(line_values[0][0]))
                property_value.append(float(line_values[1][0]))
            model = PiecewiseLinear(
                property_name, x=np.array(parameter_value), y=np.array(property_value)
            )

        assert model is not None, "Invalid property data input"
        return model



"""Provides a helper to manage materials in the Ansys ecosystem."""
import sys

if sys.version_info >= (3, 8):
    from importlib import metadata as metadata

    __version__ = metadata.version("ansys-materials-manager")

else:
    from importlib_metadata import metadata as metadata_backport

    __version__ = metadata_backport("ansys-materials-manager")["version"]



property_codes = {
    "thermal expansion coefficient x direction": "therm_exp_coeff",
    "density": "density",
    "specific heat capacity": "specific_heat",
    "molar mass": "molecular_weight",
    "thermal conductivity x direction": "thermal_conductivity",
    "viscosity": "viscosity",
    "speed of sound": "speed_of_sound",
}



"""Define a map between MAPDL and MATLM material properties."""

# todo: add more property sets and parameters to the map

# either define properties or mappings, but not both
MATML_PROPERTY_MAP = {
    "Density": {"properties": ["Density"], "mappings": {}},
    "Elasticity::Isotropic": {
        "properties": [],
        "mappings": {
            "Young's Modulus": [
                "Young's Modulus X direction",
                "Young's Modulus Y direction",
                "Young's Modulus Z direction",
            ],
            "Shear Modulus": [
                "Shear Modulus XY",
                "Shear Modulus XZ",
                "Shear Modulus YZ",
            ],
            "Poisson's Ratio": [
                "Poisson's Ratio XY",
                "Poisson's Ratio XZ",
                "Poisson's Ratio YZ",
            ],
        },
    },
    "Elasticity::Orthotropic": {
        "properties": [
            "Young's Modulus X direction",
            "Young's Modulus Y direction",
            "Young's Modulus Z direction",
            "Shear Modulus XY",
            "Shear Modulus XZ",
            "Shear Modulus YZ",
            "Poisson's Ratio XY",
            "Poisson's Ratio XZ",
            "Poisson's Ratio YZ",
        ],
        "mappings": {},
    },
    "Coefficient of Thermal Expansion::Isotropic": {
        "properties": [],
        "mappings": {
            "Coefficient of Thermal Expansion": [
                "thermal expansion coefficient x direction",
                "thermal expansion coefficient y direction",
                "thermal expansion coefficient z direction",
            ]
        },
    },
    "Coefficient of Thermal Expansion::Orthotropic": {
        "properties": [],
        "mappings": {
            "Coefficient of Thermal Expansion X direction": [
                "thermal expansion coefficient x direction"
            ],
            "Coefficient of Thermal Expansion Y direction": [
                "thermal expansion coefficient y direction"
            ],
            "Coefficient of Thermal Expansion Z direction": [
                "thermal expansion coefficient z direction"
            ],
        },
    },
    "Specific Heat": {"properties": [], "mappings": {"Specific Heat": ["Specific Heat Capacity"]}},
    "Thermal Conductivity::Isotropic": {
        "properties": [],
        "mappings": {
            "Thermal Conductivity": [
                "Thermal Conductivity X direction",
                "Thermal Conductivity Y direction",
                "Thermal Conductivity Z direction",
            ]
        },
    },
    "Thermal Conductivity::Orthotropic": {
        "properties": [
            "Thermal Conductivity X direction",
            "Thermal Conductivity Y direction",
            "Thermal Conductivity Z direction",
        ],
        "mappings": {},
    },
    "Viscosity": {"properties": ["Viscosity"], "mappings": {}},
    "Speed of Sound": {"properties": ["Speed of Sound"], "mappings": {}},
}



"""Provides the ``Material`` class."""
from typing import List, Optional

from ._models import _BaseModel
from ._models._common.constant import Constant


class Material:
    """
    Provides a wrapper class for managing a material.

    This class associates a material ID with one or more properties and nonlinear material models.
    """

    _models: List[_BaseModel]
    _id: str
    _name: str
    _uuid: str

    def __init__(
        self, material_name: str, material_id: str = None, models: List[_BaseModel] = None
    ):
        """
        Create an instance of a material.

        Optionally specify a material ID, or other properties.

        Parameters
        ----------
        material_name : str
            Name of the material.
        material_id : str, optional
            ID to associate with this material. The default is ``None``.
        models : Dict[str, _BaseModel]
            Dictionary of nonlinear material models. Models are specified with their
            model codes (from the TB command) and the model object.
        """
        self._models = []
        self._name = material_name
        self._id = material_id
        self._uuid = ""
        if models is not None:
            self.models.extend(models)
        if len(self.get_model_by_name("Strain Reference Temperature")) == 0:
            self.models.append(Constant("Strain Reference Temperature", 0))

    @property
    def name(self) -> str:
        """Name of the material."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def material_id(self) -> Optional[str]:
        """Material ID."""
        return self._id

    @material_id.setter
    def material_id(self, value: str):
        self._id = value

    @property
    def uuid(self) -> str:
        """UUID (transfer ID), which is unique."""
        return self._uuid

    @uuid.setter
    def uuid(self, value: str) -> None:
        self._uuid = value

    @property
    def models(self) -> "List[_BaseModel]":
        """Currently assigned material models."""
        return self._models

    def get_model_by_name(self, model_name: str) -> "List[_BaseModel]":
        """Get the material model or models with a given model name."""
        return [model for model in self.models if model.name.lower() == model_name.lower()]

    @property
    def reference_temperature(self) -> float:
        """Strain reference temperature for the model."""
        reference_temperature = self.get_model_by_name("Strain Reference Temperature")[0]
        assert isinstance(reference_temperature, Constant)
        return reference_temperature.value

    @reference_temperature.setter
    def reference_temperature(self, value: float) -> None:
        reference_temperature = self.get_model_by_name("Strain Reference Temperature")[0]
        assert isinstance(reference_temperature, Constant)
        reference_temperature.value = value



"""Provides the ``TableDataParser`` class."""
from typing import Dict, List, Optional

from ansys.materials.manager._models import _BaseModel
from ansys.materials.manager.util.common import TB_MATERIAL_HEADER_REGEX


class _TableDataParser:
    """Parses responses from the `TBLIST` command into individual non-linear models."""

    models: Dict[str, _BaseModel]

    def __init__(self, models: Dict[str, _BaseModel]):
        self.models = models

    @staticmethod
    def _get_tb_sections_with_id(data: str, id_: int) -> Dict[str, List[str]]:
        """
        Extract material property information for the specified material identity.

        Parameters
        ----------
        data : str
            String response from the `TBLIST` command.
        id_ : int
            Material identity to be extracted.

        Returns
        -------
        Dict[str, List[str]]
            Dictionary of relevant non-linear model sections, indexed by model name. Each section
            is split on newlines.
        """
        tb_chunks: Dict[str, List[str]] = {}
        header_lines = TB_MATERIAL_HEADER_REGEX.findall(data)
        for line in header_lines:
            if len(line) == 2 and int(line[1]) == id_:
                tb_chunks[line[0]] = []

        if len(tb_chunks) == 0:
            raise IndexError(f"Material with ID {id_} not found in data")

        material_code = None
        for line in data.splitlines():
            stripped_line = line.strip()
            match = TB_MATERIAL_HEADER_REGEX.search(stripped_line)
            if match:
                current_material_id = int(match.groups()[1])
                if current_material_id == id_:
                    material_code = match.groups()[0]
                else:
                    material_code = None
            if material_code is not None:
                tb_chunks[material_code].append(line)

        return tb_chunks

    def get_models_by_id(self, data: str, material_id: int) -> Optional[Dict[str, _BaseModel]]:
        material_model_data = self._get_tb_sections_with_id(data, material_id)
        model_data_map = {}
        for model_code, model_data in material_model_data.items():
            model_data_map[model_code] = self.deserialize_model(model_code, model_data)
        return model_data_map

    def deserialize_model(self, model_code: str, model_data: List[str]) -> _BaseModel:
        """
        Deserialize a model definition into a model object.

        Parameters
        ----------
        model_code : str
            String model code. See :meth:`ansys.mapdl.core._commands._preproc.materials.tb`.
        model_data : List[str]
            List of lines containing model data for one material and one model.

        Returns
        -------
        _BaseModel
            A model object.

        Raises
        ------
        NotImplementedError
            If a wrapper with the specified model_code does not exist.

        """
        if model_code not in self.models.keys():
            raise NotImplementedError(f"Model with key '{model_code}' is not implemented yet.")
        target_model = self.models[model_code]
        return target_model.deserialize_model(model_code, model_data)



from typing import Any, List, Tuple

import numpy as np

from ansys.materials.manager._models._common._base import _BaseModel, _FluentCore, _MapdlCore
from ansys.materials.manager._models._common._exceptions import ModelValidationException
from ansys.materials.manager._models._common._packages import SupportedPackage
from ansys.materials.manager._models._fluent.simple_properties import (
    property_codes as fluent_property_codes,
)
from ansys.materials.manager._models._mapdl.simple_properties import (
    property_codes as mapdl_property_codes,
)
from ansys.materials.manager.util.common import _chunk_data

TYPE_CHECKING = False
if TYPE_CHECKING:
    from ansys.materials.manager.material import Material  # noqa: F401


class PiecewiseLinear(_BaseModel):
    """
    Provides the piecewise linear model for a property.

    This class represents a property as a series of sample points between which the value is
    interpolated.
    """

    applicable_packages: SupportedPackage.MAPDL | SupportedPackage.FLUENT
    _name: str
    _independent_variable: np.ndarray
    _dependent_variable: np.ndarray

    def __init__(self, name: str, x: np.ndarray, y: np.ndarray) -> None:
        """
        Create a piecewise linear model for a property.

        Ensure that the two arrays ``x`` and ``y`` are both one dimensional and have the same number
        of elements.

        This property is created in the default unit system of the solver. Ensure
        that you provide values in the correct units.

        Parameters
        ----------
        name: str
            Name of the property to model as a constant.
        x: np.ndarray
            Values of the independent variable over which to model the property.
            Ensure that these values cover the range of interest for the simulation.
        y: np.ndarray
            Values of the property sampled at the points represented by the ``x`` array.
        """
        self._name = name
        self._independent_variable = x
        self._dependent_variable = y

    @property
    def name(self) -> str:
        """Name of the quantity modeled by the constant."""
        return self._name

    @property
    def x(self) -> np.ndarray:
        """Values of the ``x`` array."""
        return self._independent_variable

    @x.setter
    def x(self, value: np.ndarray) -> None:
        self._independent_variable = value

    @property
    def y(self) -> np.ndarray:
        """Values of the ``y`` array."""
        return self._dependent_variable

    @y.setter
    def y(self, value: np.ndarray) -> None:
        self._dependent_variable = value

    def write_model(self, material: "Material", pyansys_session: Any) -> None:
        """
        Write the model to MAPDL.

        This method should make some effort to validate the model state before writing.

        Parameters
        ----------
        material: Material
            Material object to associate with this model.
        pyansys_session: Any
            Configured instance of the PyAnsys session.
        """
        is_ok, issues = self.validate_model()
        if not is_ok:
            raise ModelValidationException("\n".join(issues))

        if isinstance(pyansys_session, _MapdlCore):
            self._write_mapdl(pyansys_session, material)
        elif isinstance(pyansys_session, _FluentCore):
            self._write_fluent(pyansys_session, material)
        else:
            raise TypeError(
                "This model is only supported by MAPDL and Fluent. Ensure that you have the correct"
                "type of the PyAnsys session."
            )

    def _write_mapdl(self, mapdl: "_MapdlCore", material: "Material") -> None:
        if self._independent_variable.size > 100:
            raise ValueError(
                f"MAPDL supports up to 100 points for a property. "
                f"You provided {self._independent_variable.size}."
            )
        mapdl_property_code = mapdl_property_codes[self._name.lower()]
        temp_values = self._independent_variable
        for index, chunk in enumerate(_chunk_data(temp_values)):
            mapdl.mptemp(6 * index + 1, *chunk)
        property_values = self._dependent_variable
        for index, chunk in enumerate(_chunk_data(property_values)):
            mapdl.mpdata(mapdl_property_code, material.material_id, 6 * index + 1, *chunk)

    def _write_fluent(self, fluent: "_FluentCore", material: "Material") -> None:
        fluent_property_code = fluent_property_codes[self._name.lower()]
        pass

    def validate_model(self) -> "Tuple[bool, List[str]]":
        """
        Perform pre-flight validation of the model setup.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        failures = []
        is_ok = True

        if self._name is None or self._name == "":
            failures.append("Invalid property name")
            is_ok = False
        if self._independent_variable is None or self._independent_variable.size == 0:
            failures.append("x_values is empty. Provide at least one value.")
            is_ok = False
        if self._dependent_variable is None or self._dependent_variable.size == 0:
            failures.append("y_values is empty. Provide at least one value.")
            is_ok = False
        if not is_ok:
            return is_ok, failures
        if self._independent_variable.size != self._dependent_variable.size:
            failures.append(
                "Length mismatch. x_values and y_values must have the same number of elements."
            )
            is_ok = False
        if self._independent_variable.ndim > 1:
            failures.append(
                f"x_values must have one dimension, not {self._independent_variable.ndim}."
            )
            is_ok = False
        if self._dependent_variable.ndim > 1:
            failures.append(
                f"y_values must have one dimension, not {self._dependent_variable.ndim}."
            )
            is_ok = False
        return is_ok, failures



"""Provides the ``matml_parser`` module."""
from dataclasses import dataclass
import os
from typing import Any, Dict, Union
import xml.etree.ElementTree as ET

_PATH_TYPE = Union[str, os.PathLike]

MATERIALS_ELEMENT_KEY = "Materials"
MATML_DOC_KEY = "MatML_Doc"
METADATA_KEY = "Metadata"
BULKDATA_KEY = "BulkDetails"
UNITLESS_KEY = "Unitless"
BEHAVIOR_KEY = "Behavior"
WBTRANSFER_KEY = "ANSYSWBTransferData"
MAT_TRANSFER_ID = "DataTransferID"


# Todos:
#   variable material properties with interpolation settings
#   version handling
#   support of units (exponents)


@dataclass
class Parameter:
    """Define a parameter such as density or Young's Modulus."""

    # todo: units

    name: str
    data: Any
    qualifiers: Dict


@dataclass
class PropertySet:
    """Define a PropertySet which contains one or several parameters."""

    name: str
    parameters: Dict
    qualifiers: Dict


class MatmlReader:
    """
    Parse a MATML (engineering data xml) file.

    Fills a nested dict with all the materials and their properties.
    The key of the first layer are the material names.
    The conversion into a specific format/object representation is implemented separately.

    The data can be accessed via matml_reader.materials
    """

    materials: Dict
    transfer_ids = Dict
    matml_file_path: _PATH_TYPE

    def __init__(self, file_path: _PATH_TYPE):
        """
        Create a new MATML reader object.

        Parameters
        ----------
        file_path :
            Matml (engineering data xml) file path
        """
        self.matml_file_path = file_path
        if not os.path.exists(file_path):
            raise RuntimeError(f"Cannot initialize MatmlReader {file_path}. File does not exist!")

    def _convert(self, data: str, target: str) -> Union[str, float]:
        # convert a string into a certain format (e.g. float)

        if target == "string":
            return data

        if target != "float":
            raise RuntimeError(f"unsupported format {target}. Skipped formatting {data}")

        if not data or not data.strip():
            return 0.0

        if data.count(",") > 0:
            data = data.split(",")
            return [float(v) for v in data]
        else:
            return float(data)

    def _read_metadata(self, metadata_node: Any) -> Dict:
        # Read the metadata

        data = {}
        for item in metadata_node.iter("ParameterDetails"):
            id = item.attrib["id"]
            name = item.find("Name").text
            if item.find(UNITLESS_KEY) is not None:
                unit = item.find(UNITLESS_KEY).tag
            elif item.find("Units") is not None:
                unit = item.find("Units").attrib["name"]
            else:
                raise RuntimeError(f"unhandled case {id}")

            data[id] = {"Name": name, "Units": unit}

        for item in metadata_node.iter("PropertyDetails"):
            id = item.attrib["id"]
            name = item.find("Name").text
            unit = UNITLESS_KEY

            data[id] = {"Name": name, "Units": unit}

        return data

    def _read_qualifiers(self, property_node: Any) -> Dict:
        # returns the qualifiers such as behavior, interpolation options etc.
        qualifiers = {}
        for item in property_node.findall("Qualifier"):
            qualifiers[item.attrib["name"]] = item.text
        return qualifiers

    def _read_property_sets_and_parameters(self, bulkdata: Any, metadata_dict: Dict) -> Dict:
        prop_dict = {}

        # iterate over the property sets
        for prop_data in bulkdata.findall("PropertyData"):
            property_key = prop_data.attrib["property"]
            property_name = metadata_dict[property_key]["Name"]
            prop_set_qualifiers = self._read_qualifiers(prop_data)

            parameters = {}

            # iterate over each parameter
            for parameter in prop_data.findall("ParameterValue"):
                parameter_key = parameter.attrib["parameter"]
                parameter_name = metadata_dict[parameter_key]["Name"]
                parameter_format = parameter.attrib["format"]
                param_qualifiers = self._read_qualifiers(parameter)
                data = self._convert(parameter.find("Data").text, parameter_format)

                parameters[parameter_name] = Parameter(
                    name=parameter_name, data=data, qualifiers=param_qualifiers
                )

            prop_dict[property_name] = PropertySet(
                name=property_name, qualifiers=prop_set_qualifiers, parameters=parameters
            )

        return prop_dict

    def _read_materials(self, matml_doc_node: Any, metadata_dict: Dict) -> Dict:
        # Import the material properties and parameters

        materials = {}
        for material in matml_doc_node.findall("Material"):
            bulkdata = material.find("BulkDetails")
            name = bulkdata.find("Name").text
            data = self._read_property_sets_and_parameters(bulkdata, metadata_dict)
            materials[name] = data

        return materials

    def _read_transfer_ids(self, root: ET.Element) -> int:
        # reads the material transfer IDs
        self.transfer_ids = {}
        wb_transfer_element = root.find(WBTRANSFER_KEY)
        if wb_transfer_element:
            materials_element = wb_transfer_element.find(MATERIALS_ELEMENT_KEY)
            for mat in materials_element.findall("Material"):
                mat_name = mat.find("Name").text
                transfer_id_element = mat.find(MAT_TRANSFER_ID)

                if not mat_name in self.materials.keys():
                    raise RuntimeError(f"Transfer ID could not be set for material {mat_name}")

                self.transfer_ids[mat_name] = transfer_id_element.text

        return len(self.transfer_ids)

    def parse_matml(self) -> bool:
        """Read MATML (engineering data XML) file.

        Output can be consumed via matml_reader.materials or
        matml_reader.get_material(name).

        Returns the number of imported materials.
        """
        tree = ET.parse(self.matml_file_path)
        root = tree.getroot()
        materials_node = root.find(MATERIALS_ELEMENT_KEY)
        if not materials_node:
            raise RuntimeError(
                "Materials node not found. Please check if this is a valid MATML file."
            )

        matml_doc_node = materials_node.find(MATML_DOC_KEY)
        if not matml_doc_node:
            raise RuntimeError("MATML node not found. Please check if this is a valid MATML file.")

        metadata_node = matml_doc_node.find(METADATA_KEY)
        if not metadata_node:
            raise RuntimeError(
                "Metadata node not found. Please check if this is a valid MATML file."
            )
        metadata_dict = self._read_metadata(metadata_node)

        self.materials = self._read_materials(matml_doc_node, metadata_dict)

        self._read_transfer_ids(root)
        return len(self.materials)

    def get_material(self, name: str) -> Dict:
        """Return a certain material."""
        if not name in self.materials.keys():
            available_keys = ", ".join(self.materials.keys())
            raise RuntimeError(
                f"Material {name} does not exist. Available materials are {available_keys}"
            )

        return self.materials[name]



from abc import ABCMeta, abstractmethod
from typing import Any, List, Tuple

try:
    from ansys.mapdl.core.mapdl import _MapdlCore
except ImportError:
    _MapdlCore = type(None)

try:
    from ansys.fluent.core import Fluent as _FluentCore
except ImportError:
    _FluentCore = type(None)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from ansys.materials.manager._models._common._packages import SupportedPackage  # noqa: F401
    from ansys.materials.manager.material import Material  # noqa: F401


class _BaseModel(metaclass=ABCMeta):
    """
    Provides the base class that all nonlinear material models must inherit from.

    This class allows the Material Manager to dynamically discover available models and to dispatch
    deserialization calls to the appropriate model class.
    """

    applicable_packages: "SupportedPackage"

    @property
    @abstractmethod
    def name(self) -> str:
        """Get the name of the model.

        For complex nonlinear models, this is the name of the model. For simple models, this
        can be set and should reflect the property being modeled.
        """
        ...

    @abstractmethod
    def write_model(self, material: "Material", pyansys_session: Any) -> None:
        """
        Write the model to MAPDL.

        This method should make some effort to validate the model state before writing.

        Parameters
        ----------
        material: Material
            Material object to associate this model with.
        pyansys_session: Any
            Supported PyAnsys product session. Only PyMAPDL and PyFluent are
            supported currently.
        """
        ...

    @abstractmethod
    def validate_model(self) -> "Tuple[bool, List[str]]":
        """
        Perform pre-flight validation of the model setup.

        This method should not perform any calls to the MAPDL process.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        ...



"""Provides a function to convert MatML entries into Material objects."""
from typing import Dict, Sequence

from ansys.materials.manager._models import Constant
from ansys.materials.manager.material import Material
from ansys.materials.manager.util.matml.matml_property_map import MATML_PROPERTY_MAP


def convert_matml_materials(
    materials_dict: Dict, transfer_ids: Dict, index_offset: int
) -> Sequence[Material]:
    """
    Convert a list of materials into Material objects.

    Parameters
    ----------
    materials_dict:
        dict of raw material data from a matml import
    transfer_ids:
        dict of material names and unique transfer ids
    index_offset:
        int to offset the material id (number) to avoid conflicts with already existing materials
    Returns a list of Material objects
    """
    materials = []

    global_material_index = 1 + index_offset
    # loop over the materials
    for mat_id, material_data in materials_dict.items():

        models = []
        # loop over the defined property sets
        for propset_name, property_set in material_data.items():

            if "Behavior" in property_set.qualifiers.keys():
                propset_name += "::" + property_set.qualifiers["Behavior"]

            # check if the Material object supports this property set
            if propset_name in MATML_PROPERTY_MAP.keys():
                parameter_map = MATML_PROPERTY_MAP[propset_name]

                for property_name in parameter_map["properties"]:
                    param = property_set.parameters[property_name]
                    value = param.data
                    if isinstance(value, Sequence):
                        if len(value) > 1:
                            raise RuntimeError(
                                f"Only constant material properties are supported ATM. "
                                f"Value of `{property_name}` is `{value}`"
                            )
                        value = value[0]
                    models.append(Constant(property_name, value))
                for matml_key, mapped_properties in parameter_map["mappings"].items():
                    param = property_set.parameters[matml_key]
                    value = param.data
                    if isinstance(value, Sequence):
                        if len(value) > 1:
                            raise RuntimeError(
                                f"Only constant material properties are supported ATM. "
                                f"Value of `{matml_key}` is `{value}`"
                            )
                        value = value[0]
                    for mapped_property in mapped_properties:
                        models.append(Constant(mapped_property, value))

        mapdl_material = Material(
            material_name=mat_id, material_id=global_material_index, models=models
        )

        if mat_id in transfer_ids.keys():
            mapdl_material.uuid = transfer_ids[mat_id]

        materials.append(mapdl_material)

        global_material_index += 1

    return materials



property_codes = {
    "young's modulus x direction": "EX",
    "young's modulus y direction": "EY",
    "young's modulus z direction": "EZ",
    "secant thermal expansion coefficient x direction": "ALPX",
    "secant thermal expansion coefficient y direction": "ALPY",
    "secant thermal expansion coefficient z direction": "ALPZ",
    "instantaneous thermal expansion coefficient x direction": "CTEX",
    "instantaneous thermal expansion coefficient y direction": "CTEY",
    "instantaneous thermal expansion coefficient z direction": "CTEZ",
    "strain reference temperature": "REFT",
    "poisson's ratio xy": "PRXY",
    "poisson's ratio yz": "PRYZ",
    "poisson's ratio xz": "PRXZ",
    "shear modulus xy": "GXY",
    "shear modulus yz": "GYZ",
    "shear modulus xz": "GXZ",
    "coefficient of friction": "MU",
    "density": "DENS",
    "specific heat capacity": "C",
    "enthalpy": "ENTH",
    "thermal conductivity x direction": "KXX",
    "thermal conductivity y direction": "KYY",
    "thermal conductivity z direction": "KZZ",
    "convection coefficient": "HF",
    "emissivity": "EMIS",
    "heat generation rate": "QRATE",
    "viscosity": "VISC",
    "speed of sound": "SONC",
    "electrical resistivity x direction": "RSVX",
    "electrical resistivity y direction": "RSVY",
    "electrical resistivity z direction": "RSVZ",
    "electric relative permittivity x direction": "PERX",
    "electric relative permittivity y direction": "PERY",
    "electric relative permittivity z direction": "PERZ",
    "magnetic relative permeability x direction": "MURX",
    "magnetic relative permeability y direction": "MURY",
    "magnetic relative permeability z direction": "MURZ",
    "magnetic coercive force x direction": "MGXX",
    "magnetic coercive force y direction": "MGYY",
    "magnetic coercive force z direction": "MGZZ",
    "dielectric loss tangent": "LSST",
    "seebeck coefficient x direction": "SBKX",
    "seebeck coefficient y direction": "SBKY",
    "seebeck coefficient z direction": "SBKZ",
}



from typing import Any, List, Optional, Tuple

import numpy as np

from ansys.materials.manager._models._common._base import _BaseModel, _FluentCore, _MapdlCore
from ansys.materials.manager._models._common._exceptions import ModelValidationException
from ansys.materials.manager._models._common._packages import SupportedPackage
from ansys.materials.manager._models._fluent.simple_properties import (
    property_codes as fluent_property_codes,
)
from ansys.materials.manager._models._mapdl.simple_properties import (
    property_codes as mapdl_property_codes,
)
from ansys.materials.manager.util.common import _chunk_data

TYPE_CHECKING = False
if TYPE_CHECKING:
    from ansys.materials.manager.material import Material  # noqa: F401


class Polynomial(_BaseModel):
    """
    Provides the polynomial model for a property.

    This class represents a property as a polynomial with a series of coefficients.
    These are stored in ascending powers of the parameters.
    """

    applicable_packages: SupportedPackage.MAPDL | SupportedPackage.FLUENT
    _name: str
    _coefficients: np.ndarray
    _sample_points: np.ndarray

    def __init__(
        self, name: str, coefficients: np.ndarray, sample_points: Optional[np.ndarray] = None
    ) -> None:
        """
        Create a polynomial model for a property.

        This model represents an nth order polynomial. The coefficients are stored
        and entered in ascending order. For some packages, the polynomial is evaluated
        and then interpolated. In these cases sample points may be provided and should
        cover the range of interest.

        For example, providing ``[3., 4., 1.]`` represents this equation:

        .. math::

        f(x) = 3 + 4x + x^{2}

        This property is created in the default unit system of the solver. Ensure
        that you provide values in the correct units.

        Parameters
        ----------
        name: str
            Name of the property to model as a constant.
        coefficients: np.ndarray
            Values of the coefficients for the polynomial model, provided in ascending
            powers of the parameter.
        sample_points: Optional[np.ndarray]
            Values of the parameter to use for interpolation, which are required for some packages.
        """
        self._name = name
        self._coefficients = coefficients
        self._sample_points = sample_points

    @property
    def name(self) -> str:
        """Name of the quantity modeled by this constant."""
        return self._name

    @property
    def coefficients(self) -> np.ndarray:
        """Values of the ``x`` array."""
        return self._coefficients

    @coefficients.setter
    def coefficients(self, value: np.ndarray) -> None:
        self._coefficients = value

    def write_model(self, material: "Material", pyansys_session: Any) -> None:
        """
        Write this model to MAPDL.

        This method Should make some effort to validate the model state before writing.

        Parameters
        ----------
        material: Material
            Material object to associate with this model.
        pyansys_session: Any
            Configured instance of the PyAnsys session.
        """
        is_ok, issues = self.validate_model()
        if not is_ok:
            raise ModelValidationException("\n".join(issues))

        if isinstance(pyansys_session, _MapdlCore):
            self._write_mapdl(pyansys_session, material)
        elif isinstance(pyansys_session, _FluentCore):
            self._write_fluent(pyansys_session, material)
        else:
            raise TypeError(
                "This model is only supported by MAPDL and Fluent. Ensure that you have the correct"
                "type of the PyAnsys session."
            )

    def _write_mapdl(self, mapdl: "_MapdlCore", material: "Material") -> None:
        if self._coefficients.size > 5:
            raise ValueError(
                "MAPDL supports up to 5 coefficients, corresponding to a fourth order polynomial,"
                f"you provided {self._coefficients.size}"
            )
        if self._sample_points is not None:
            if self._sample_points.size > 100:
                raise ValueError(
                    "MAPDL supports up to 100 sample points for interpolation. "
                    f"You provided {self._sample_points.size}."
                )
            for index, chunk in enumerate(_chunk_data(self._sample_points)):
                mapdl.mptemp(6 * index + 1, *chunk)
        mapdl_property_code = mapdl_property_codes[self._name.lower()]
        mapdl.mp(mapdl_property_code, material.material_id, *self._coefficients)

    def _write_fluent(self, fluent: "_FluentCore", material: "Material") -> None:
        fluent_property_code = fluent_property_codes[self._name.lower()]
        pass

    def validate_model(self) -> "Tuple[bool, List[str]]":
        """
        Perform pre-flight validation of the model setup.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        failures = []
        is_ok = True

        if self._name is None or self._name == "":
            failures.append("Invalid property name")
            is_ok = False
        if self._coefficients is None or self._coefficients.size == 0:
            failures.append("Coefficients is empty. Provide at least one value.")
            is_ok = False
        if not is_ok:
            return is_ok, failures
        if self._coefficients.ndim > 1:
            failures.append(f"Coefficients must have one dimension, not {self._coefficients.ndim}.")
            is_ok = False
        if self._sample_points is not None and self._sample_points.ndim > 1:
            failures.append(
                f"Sample points must have one dimension, not {self._sample_points.ndim}."
            )
            is_ok = False
        return is_ok, failures



"""Provides the ``MaterialManager`` class."""
import inspect
from typing import Any, Dict

import ansys.materials.manager._models as models
from ansys.materials.manager._models._common._base import _FluentCore, _MapdlCore

from ._models import _BaseModel
from .material import Material
from .util.mapdl.mapdl_reader import read_mapdl


class MaterialManager:
    """
    Manage material creation, assignment, and other management tasks.

    This class is the main entry point for the Pythonic material management interface.
    """

    model_type_map: Dict[str, models._BaseModel] = {}
    _client: Any

    def __init__(self, pyansys_client: Any):
        """
        Create a ``MaterialManager`` object ready for use.

        Parameters
        ----------
        pyansys_client : Any
            Valid instance of a PyAnsys client. Only PyMAPDL and PyFluent are
            supported currently.
        """
        self._client = pyansys_client
        # response = inspect.getmembers(models, self.__is_subclass_predicate)
        # model_classes: List[models._BaseModel] = [tple[1] for tple in response]
        # for class_ in model_classes:
        #     supported_model_codes = class_.model_codes
        #     for model_code in supported_model_codes:
        #         self.model_type_map[model_code] = class_

    @staticmethod
    def __is_subclass_predicate(obj: object) -> bool:
        """
        Determine if an object is a strict subclass of the :obj:`models._BaseModel` class.

        Parameters
        ----------
        obj : object
            Any Python object.

        Returns
        -------
        bool
            ``True`` if the object is strictly a subclass of the :obj:`models._BaseModel`
            class, ``False`` otherwise.
        """
        return (
            isinstance(obj, type)
            and issubclass(obj, models._BaseModel)
            and not inspect.isabstract(obj)
        )

    def write_material(self, material: Material) -> None:
        """
        Write a material to the solver.

        Parameters
        ----------
        material : Material
            Material object to write to solver.
        """
        for model in material.models:
            assert isinstance(model, _BaseModel)
            model.write_model(material, self._client)

    def read_materials_from_session(self) -> Dict[str, Material]:
        """
        Given a PyAnsys session, return the materials present.

        This method only supports PyMAPDL currently.

        Returns
        -------
        Dict[str, Material]
            Materials in the current session, indexed by an ID. For MAPDL, this is the material
            ID. For Fluent, this is the material name.
        """
        if isinstance(self._client, _MapdlCore):
            return self._read_mapdl()
        elif isinstance(self._client, _FluentCore):
            return self._read_fluent()

    def _read_mapdl(self) -> Dict[str, Material]:
        return read_mapdl(self._client)

    def _read_fluent(self) -> Dict[str, Material]:
        return []



"""Provides the ``mapdl_reader`` module."""
from typing import Dict

from ansys.mapdl.core.mapdl import _MapdlCore

from ansys.materials.manager.material import Material

from .mpdata_parser import MP_MATERIAL_HEADER_REGEX, _MaterialDataParser


def read_mapdl(mapdl: _MapdlCore) -> Dict[str, Material]:
    """
    Read materials from a provided MAPDL session.

    Returns them indexed by the material ID.

    Parameters
    ----------
    mapdl : _MapdlCore
        Active pyMAPDL session.

    Returns
    -------
    Dict[str, Material]
        Materials currently active in the MAPDL session, indexed by their material ID.
    """
    materials = []
    data = mapdl.mplist()
    material_ids = list(MP_MATERIAL_HEADER_REGEX.findall(data))
    for material_id in material_ids:
        material_properties = _MaterialDataParser.parse_material(data, int(material_id))
        materials.append(Material("", material_id, models=material_properties))
    return {
        material.material_id: material for material in materials if material.material_id is not None
    }



"""Provides the ``common`` module."""
from itertools import islice
import re
from typing import Generator, Iterable, List, Union

import numpy as np

MP_MATERIAL_HEADER_REGEX = re.compile(r"MATERIAL NUMBER\s+([\d]+)")
TB_MATERIAL_HEADER_REGEX = re.compile(r"\(([A-Z]+)\) Table For Material\s+([\d]+)[^\n]*")
FLOAT_VALUE_REGEX = re.compile(r"(-?\d+\.\d*([Ee][+-]\d+)?)")
INTEGER_VALUE_REGEX = re.compile(r"(-?\d+)")
MATRIX_LABEL_REGEX = re.compile(r"(\w\s?\d{1,2})")

model_type = Union[float, np.ndarray]


def _chunk_data(data: Iterable) -> Generator[List, None, None]:
    """
    Split an iterable into chunks of size six lazily.

    Parameters
    ----------
    data : Iterable
        Input data.

    Returns
    -------
    Generator[List]
        Generator that will return the input data in lists of length 6 or less.
    """
    data_iterator = iter(data)
    piece = list(islice(data_iterator, 6))
    while piece:
        yield piece
        piece = list(islice(data_iterator, 6))


def _chunk_lower_triangular_matrix(
    matrix: np.ndarray,
) -> Generator[Iterable[float], None, None]:
    """
    Convert a lower-triangular square matrix into a chunked vector.

    This is intended for use with symmetric matrices, where the upper half can be ignored.

    Parameters
    ----------
    matrix : np.ndarray
        Input square matrix.

    Returns
    -------
    Generator[Iterable[float]]
        Chunked vector with coefficients in the lower half-matrix. E.g. D11, D12, D22, etc.

    Notes
    -----
    If the input matrix is non-symmetric then the upper half-matrix will be lost.
    """
    vals = []
    if np.ndim(matrix) != 2:
        raise ValueError("Input matrix must be two dimensional.")
    size = matrix.shape
    if size[0] != size[1]:
        raise ValueError("Input matrix must be square.")
    for row_index in range(0, matrix.shape[0]):
        for col_index in range(0, row_index + 1):
            vals.append(matrix[row_index][col_index])
    return _chunk_data(vals)


def fill_upper_triangular_matrix(vector: List[float]) -> np.ndarray:
    """
    Convert a vector of coefficients into a full matrix.

    Generates a symmetric, square matrix.

    Parameters
    ----------
    vector : List[float]
        Coefficients of the lower half-matrix. E.g. D11, D12, D22, etc.

    Returns
    -------
    np.ndarray
        Square symmetric matrix.

    Raises
    ------
    ValueError
        If the length of the input vector is not a triangular number.
    """
    size_x = (np.sqrt(8 * len(vector) + 1) - 1) / 2
    if not np.isclose(int(size_x), size_x):
        raise ValueError(
            f"vector does not appear to be a valid lower-triangular matrix in flat form "
            f"(size {len(vector)}) is not a triangular number"
        )
    size_x = int(size_x)
    output = np.zeros((size_x, size_x))
    output[np.triu_indices(output.shape[0], k=0)] = vector
    return output + output.T - np.diag(np.diag(output))



from enum import Enum
from typing import Any, List, Optional, Tuple, Union

import numpy as np

from ansys.materials.manager._models._common._base import _BaseModel
from ansys.materials.manager._models._common._exceptions import ModelValidationException
from ansys.materials.manager._models._common._packages import SupportedPackage
from ansys.materials.manager.util.common import (
    FLOAT_VALUE_REGEX,
    MATRIX_LABEL_REGEX,
    _chunk_lower_triangular_matrix,
    fill_upper_triangular_matrix,
)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from ansys.materials.manager._models._common._base import _MapdlCore  # noqa: F401
    from ansys.materials.manager.material import Material  # noqa: F401


class ElasticityMode(Enum):
    """Determines which kind of coefficients are used in the model."""

    # Indicates that the model coefficients have units of stiffness, for example GPa
    STIFFNESS = 1

    # Indicates that the model coefficients have units of compliance (inverse stiffness), for
    # example GPa^-1
    COMPLIANCE = 2


class AnisotropicElasticity(_BaseModel):
    r"""
    Provides the anisotropic elasticity model.

    The anisotropic elasticity model defines different elastic coefficients
    for each coordinate axis. This model can be used with plane and solid elements.
    The elastic coefficient matrix (D) is specified as one or up to six NumPy arrays,
    allowing temperature dependence to be modeled.

    The elastic coefficient matrix is defined as a 2 x n_dimensions array:

    .. math::
          \begin{matrix}
            D_{11}\\
            D_{21} & D_{22}\\
            D_{31} & D_{32} & D_{33}\\
            D_{41} & D_{42} & D_{43} & D_44}
          \end{matrix}

    This can either be specified in "stiffness" form, with units of force/area, or in "compliance"
    form with the inverse unit.

    Notes
    -----
    This model wraps the APDL "ELAS" and "ANEL" models in both their forms. If one temperature is
    provided, the "ELAS" model is used, with either the "AELS" or "AELF" TBOPT. Otherwise,
    the "ANEL" model is used.
    """

    _n_dimensions: int
    _coefficient_type: ElasticityMode
    _coefficients: np.ndarray
    _temperature: np.ndarray

    model_codes = ("ANEL", "ELAS")
    name = "Anisotropic Elasticity"
    applicable_packages = SupportedPackage.MAPDL

    def __init__(
        self,
        n_dimensions: int,
        coefficient_type: ElasticityMode,
        coefficients: Optional[np.ndarray] = None,
        temperature: Union[None, float, np.ndarray] = None,
    ) -> None:
        """
        Create an anisotropic elasticity model.

        Parameters
        ----------
        n_dimensions: int
            Number of dimensions for the model. The value must be either ``2`` or ``3``.
        coefficient_type: ElasticityMode
            Type of elasticity mode that the model uses, which is either stiffness or
            compliance coefficients.
        coefficients: np.ndarray
            Model coefficients, either one 2*n_dims x 2*n_dims array of the model coefficients or,
            if the model is temperature dependent, up to 6 x 2*n_dims x 2*n_dims array of the model
            coefficients at temperature.
        temperature: Union[float, np.ndarray]
            Either a single temperature at which to apply the model or an array of up to
            six temperatures if the model is temperature dependent. If multiple temperatures are
            defined, ensure that they cover the range of anticipated temperatures at which the model
            is applied.
        """
        if n_dimensions not in [2, 3]:
            raise ValueError("n_dimensions must be an integer value of 2 or 3.")
        self._n_dimensions = n_dimensions
        self._coefficient_type = coefficient_type
        self._temperature = np.array([], dtype=float)
        self._coefficients = np.array([], dtype=float)
        if temperature is not None:
            if isinstance(temperature, (float, int)):
                self._temperature = np.array([temperature], dtype=float)
            else:
                self._temperature = temperature
        if coefficients is not None:
            self._coefficients = coefficients

    def __repr__(self) -> str:
        """Apparently this requires a docstring?."""
        return (
            f"<AnisotropicElasticity "
            f"n_dimensions={self._n_dimensions}, "
            f"temperature_count={len(self._temperature)}, "
            f"mode={self._coefficient_type.name}>"
        )

    @property
    def name(self) -> str:
        """Name of the model."""
        return self.name

    @property
    def coefficients(self) -> np.ndarray:
        """Current coefficient array for the model."""
        return self._coefficients

    @coefficients.setter
    def coefficients(self, value: np.ndarray):
        self._coefficients = value

    @property
    def temperature(self) -> np.ndarray:
        """Temperature defined for the model."""
        return self._temperature

    @temperature.setter
    def temperature(self, value: Union[float, np.ndarray]):
        if isinstance(value, (float, int)):
            self._temperature = np.array([value], dtype=float)
        else:
            self._temperature = value

    @classmethod
    def deserialize_model(cls, model_code: str, model_data: List[str]) -> "AnisotropicElasticity":
        """
        Convert output from the MAPDL command into an object representing the model.

        The input should be a section of output referring to one model from one material.

        Parameters
        ----------
        model_code: str
            String model code, either "ELAS" or "ANEL".
        model_data: List[str]
            Lines from MAPDL output corresponding to this model for one material.

        Returns
        -------
        AnisotropicElasticity
            Wrapper for the underlying MAPDL material model.

        Notes
        -----
        Depending on the type of the underlying model, the parameters of the returned
        `AnisotropicElasticity` model vary, but this class is returned for either
        "ELAS" or "ANEL" material models.
        """
        assert model_code in cls.model_codes, f"Invalid model code ({model_code}) provided."
        header_row_index = 0
        for index, line in enumerate(model_data):
            if line.strip().startswith("Temps"):
                header_row_index = index
                break
        if model_code == "ANEL":
            n_dim, temps, coeffs = cls.deserialize_anel_data(
                model_data[header_row_index : header_row_index + 22]
            )
            mode = ElasticityMode.STIFFNESS
            for line in model_data[header_row_index + 23 :]:
                if line.strip().startswith("Flexibility"):
                    mode = ElasticityMode.COMPLIANCE
                    break
            return cls(n_dim, mode, coeffs, temps)

        else:
            n_dim, temp, coeffs = cls.deserialize_elas_data(
                model_data[header_row_index : header_row_index + 22]
            )
            if "STIFFNESS" in model_data[header_row_index - 2]:
                mode = ElasticityMode.STIFFNESS
            else:
                mode = ElasticityMode.COMPLIANCE
            return cls(n_dim, mode, coeffs, temp)

    @staticmethod
    def deserialize_anel_data(
        model_data: List[str],
    ) -> Tuple[int, np.ndarray, np.ndarray]:
        """
        Deserialize the first section of data returned by calling `TBLIST` with an "ANEL" model.

        The first row contains the temperatures at which the model is applied, and subsequent
        rows contain each coefficient value at each specified temperature.

        Parameters
        ----------
        model_data: List[str]
            Lines from MAPDL output corresponding to the model coefficients and measured
            temperatures.
        """
        temp_values = [float(match[0]) for match in FLOAT_VALUE_REGEX.findall(model_data[0])]
        matrix_data = AnisotropicElasticity.read_matrix(model_data[1:])
        coeffs = []
        for temp_index, temp_value in enumerate(temp_values):
            data_at_temp = [row[temp_index + 1] for row in matrix_data]
            if np.allclose(data_at_temp[10:], 0):
                data_at_temp = data_at_temp[0:10]
            coeffs.append(fill_upper_triangular_matrix(data_at_temp))
        if all([matrix.size == 16 for matrix in coeffs]):
            ndim = 2
        else:
            ndim = 3
        coeff_np = np.empty((len(temp_values), 2 * ndim, 2 * ndim))
        for index, matrix in enumerate(coeffs):
            coeff_np[index, 0 : 2 * ndim, 0 : 2 * ndim] = matrix
        return ndim, np.asarray(temp_values, dtype=float), coeff_np

    @staticmethod
    def deserialize_elas_data(model_data: List[str]) -> Tuple[int, float, np.ndarray]:
        """
        Deserialize the first section of data returned by calling `TBLIST` with an "ELAS" model.

        The first row contains the temperature at which the model is applied, and subsequent
        rows contain each coefficient value.

        Parameters
        ----------
        model_data: List[str]
            Lines from MAPDL output corresponding to the model coefficients and measured
            temperature.
        """
        temp_values = float(FLOAT_VALUE_REGEX.findall(model_data[0])[0][0])
        matrix_data = AnisotropicElasticity.read_matrix(model_data[1:])
        data_at_temp = [row[1] for row in matrix_data]
        if np.allclose(data_at_temp[10:], 0):
            data_at_temp = data_at_temp[0:10]
            ndim = 2
        else:
            ndim = 3
        coeffs = fill_upper_triangular_matrix(data_at_temp)
        return ndim, temp_values, coeffs

    @staticmethod
    def read_matrix(model_data: List[str]) -> List[Tuple]:
        """
        Read a matrix from a list of strings.

        Iterate through a provided list of strings and extract, if present, a valid matrix element
        label and any subsequent floating point values.

        Parameters
        ----------
        model_data: List[str]
            Matrix coefficient section from the output of a `TBLIST` command. Any rows that do
            not begin with a label are ignored. otherwise, each row is deserialized into a tuple
            with the string label and any associated float values.
        """
        values = []
        for row in model_data:
            label = MATRIX_LABEL_REGEX.search(row)
            if label:
                current_values = FLOAT_VALUE_REGEX.findall(row)
                values.append((label.groups()[0], *(float(value[0]) for value in current_values)))
        return values

    def write_model(self, pyansys_session: Any, material: "Material") -> None:
        """
        Write the model to MAPDL.

        This method performs some pre-flight verification and writes the correct model for the
        provided values of coefficients and temperatures.

        If no temperature value is specified for the model, the current reference temperature
        for the material is used.

        Parameters
        ----------
        pyansys_session: _MapdlCore
            Configured instance of PyMAPDL.
        material: Material
            Material object to associate with this model.
        """
        if not isinstance(pyansys_session, _MapdlCore):
            raise TypeError(
                "This model is only supported by MAPDL and Fluent. Ensure that you have the correct"
                "type of the PyAnsys session."
            )

        is_ok, issues = self.validate_model()
        if not is_ok:
            raise ModelValidationException("\n".join(issues))

        if self._temperature is None:
            self._temperature = np.array(material.reference_temperature, dtype=float)

        if self._temperature.size == 1:
            # Write ELASTIC model
            ntemp = 1
            lab = "ELASTIC"
            if self._coefficient_type == ElasticityMode.STIFFNESS:
                tbopt = "AELS"
            else:
                tbopt = "AELF"
            self._coefficients = np.expand_dims(self._coefficients, 0)
        else:
            # Write ANEL model
            ntemp = self._temperature.size
            lab = "ANEL"
            if self._coefficient_type == ElasticityMode.STIFFNESS:
                tbopt = "0"
            else:
                tbopt = "1"

            # Ensure temperatures and coefficients are sorted
            sort_order = np.argsort(self._temperature)
            self._temperature = self._temperature[sort_order]
            self._coefficients = self._coefficients[sort_order, :, :]

        # Write table specification
        pyansys_session.tb(lab, material.material_id, ntemp, tbopt=tbopt)

        for temp_index, temp_val in enumerate(self._temperature):
            pyansys_session.tbtemp(temp_val)
            for chunk_index, data_chunk in enumerate(
                _chunk_lower_triangular_matrix(self._coefficients[temp_index])
            ):
                pyansys_session.tbdata(6 * chunk_index + 1, *data_chunk)

    def validate_model(self) -> Tuple[bool, List[str]]:
        """
        Validate some aspects of the model before attempting to write to MAPDL.

        This method performs these actions:

        * Validates that the number of provided temperatures match the size of the first dimension
           of the coefficient array.
        * Validates that the coefficient array is either two or three dimensional.
        * Validates that no more than six temperature samples are provided.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        coefficient_shape = self._coefficients.shape
        coef_matrix_count = None
        validation_errors = []
        is_valid = True

        if len(coefficient_shape) == 2:
            coef_matrix_count = 1
        elif len(coefficient_shape) == 3:
            coef_matrix_count = coefficient_shape[0]
        else:
            is_valid = False
            validation_errors.append("Invalid dimension of coefficients array. It must be 2 or 3.")

        # Check that size of temperature matches 3rd dimension of coefficients
        if coef_matrix_count and coef_matrix_count != self._temperature.size:
            is_valid = False
            validation_errors.append(
                f"Inconsistent number of temperature values ({self._temperature.size}) "
                f"and coefficient values ({coef_matrix_count})."
            )

        if self._temperature.size > 6:
            is_valid = False
            validation_errors.append("This model supports a maximum of six temperature values.")

        return is_valid, validation_errors



class ModelValidationException(Exception):
    """
    Provides the exception thrown if pre-flight verification fails.

    Throwing an exception indicates that the provided model state is invalid.
    """

    def __init__(self, message: str):
        """
        Create an instance of the ``ModelValidationException`` class.

        Parameters
        ----------
        message: str
            Text to display for the exception.
        """
        super().__init__(message)



from ansys.materials.manager._models._common._base import _BaseModel
from ansys.materials.manager._models._common._exceptions import ModelValidationException
from ansys.materials.manager._models._common.constant import Constant
from ansys.materials.manager._models._common.piecewise_linear import PiecewiseLinear
from ansys.materials.manager._models._common.polynomial import Polynomial
from ansys.materials.manager._models._mapdl.anisotropic_elasticity import (
    AnisotropicElasticity,
    ElasticityMode,
)



from typing import Any, List, Tuple

from ansys.materials.manager._models._common._base import _BaseModel, _FluentCore, _MapdlCore
from ansys.materials.manager._models._common._exceptions import ModelValidationException
from ansys.materials.manager._models._common._packages import SupportedPackage
from ansys.materials.manager._models._fluent.simple_properties import (
    property_codes as fluent_property_codes,
)
from ansys.materials.manager._models._mapdl.simple_properties import (
    property_codes as mapdl_property_codes,
)

TYPE_CHECKING = False
if TYPE_CHECKING:
    from ansys.materials.manager.material import Material  # noqa: F401


class Constant(_BaseModel):
    """Represents a constant property value in a solver."""

    applicable_packages: SupportedPackage.MAPDL | SupportedPackage.FLUENT
    _name: str
    _value: float

    def __init__(self, name: str, value: float) -> None:
        """
        Create a constant property value.

        This property is created in the default unit system of the solver. Ensure
        that you provide the value in the correct units.

        Parameters
        ----------
        name: str
            Name of the property to model as a constant.
        value: float
            Value of the constant property.
        """
        self._name = name
        self._value = value

    @property
    def name(self) -> str:
        """Name of the quantity modeled by the constant."""
        return self._name

    @property
    def value(self) -> float:
        """Constant value of the quantity."""
        return self._value

    @value.setter
    def value(self, value: float) -> None:
        self._value = value

    def write_model(self, material: "Material", pyansys_session: Any) -> None:
        """
        Write this model to MAPDL.

        This method should make some effort to validate the model state before writing.

        Parameters
        ----------
        material: Material
            Material object to associate with this model.
        pyansys_session: Any
            Configured instance of the PyAnsys session.
        """
        is_ok, issues = self.validate_model()
        if not is_ok:
            raise ModelValidationException("\n".join(issues))

        if isinstance(pyansys_session, _MapdlCore):
            self._write_mapdl(pyansys_session, material)
        elif isinstance(pyansys_session, _FluentCore):
            self._write_fluent(pyansys_session, material)
        else:
            raise TypeError(
                "This model is only supported by MAPDL and Fluent. Ensure that you have the correct"
                "type of the PyAnsys session."
            )

    def _write_mapdl(self, mapdl: "_MapdlCore", material: "Material") -> None:
        mapdl_property_code = mapdl_property_codes[self._name.lower()]
        mapdl.mp(mapdl_property_code, material.material_id, self._value)

    def _write_fluent(self, fluent: "_FluentCore", material: "Material") -> None:
        try:
            fluent_property_code = fluent_property_codes[self._name.lower()]
            property_state = {fluent_property_code: {"option": "constant", "value": self._value}}
            fluent.setup.materials.fluid[material.name] = property_state
        except (RuntimeError, KeyError):
            pass

    def validate_model(self) -> "Tuple[bool, List[str]]":
        """
        Perform pre-flight validation of the model setup.

        Returns
        -------
        Tuple
            First element is Boolean. ``True`` if validation is successful. If ``False``,
            the second element contains a list of strings with more information.
        """
        failures = []
        is_ok = True

        if self._name is None or self._name == "":
            failures.append("Invalid property name")
            is_ok = False
        if self._value is None:
            failures.append("Value cannot be None")
            is_ok = False
        return is_ok, failures



"""Provides the ``MatmlWriter`` class."""
import os
import sys
from typing import BinaryIO, Dict, Sequence, Union
import xml.etree.ElementTree as ET

from ansys.materials.manager.material import Material

from .matml_parser import (
    BULKDATA_KEY,
    MATERIALS_ELEMENT_KEY,
    MATML_DOC_KEY,
    METADATA_KEY,
    UNITLESS_KEY,
    WBTRANSFER_KEY,
)
from .matml_property_map import MATML_PROPERTY_MAP

_PATH_TYPE = Union[str, os.PathLike]

ROOT_ELEMENT = "EngineeringData"
VERSION = "18.0.0.60"
VERSION_DATE = "29.08.2016 15:02:00"


class MatmlWriter:
    """
    Exports a list of MAPDL materials to an engineering data XML file.

    Examples
    --------
    > writer = MatmlWriter(materials)
    > writer.export('engineering_data.xml')
    """

    _materials: Sequence[Material]
    _metadata_property_sets: Dict
    _metadata_parameters: Dict

    def __init__(self, materials: Sequence[Material]):
        """Construct a Matml writer."""
        self._materials = materials
        self._metadata_property_sets = {}
        self._metadata_parameters = {}

    def _add_parameters(self, property_element: ET.Element, material: Material, parameters: Dict):
        # add the parameters of a property set to the tree
        for mat_key, matml_key in parameters.items():
            if matml_key in self._metadata_parameters.keys():
                para_key = self._metadata_parameters[matml_key]
            else:
                index = len(self._metadata_parameters) + 1
                para_key = f"pa{index}"
                self._metadata_parameters[matml_key] = para_key

            param_element = ET.SubElement(
                property_element, "ParameterValue", {"format": "float", "parameter": para_key}
            )
            data_element = ET.SubElement(param_element, "Data")
            data_element.text = str(material.get_model_by_name(mat_key)[0].value)
            qualifier_element = ET.SubElement(param_element, "Qualifier", {"name": "Variable Type"})
            qualifier_element.text = "Dependent"

    def _add_property_set(
        self,
        bulkdata_element: ET.Element,
        material: Material,
        property_set_name: str,
        parameter_map: Dict,
        behavior: str,
    ):
        """Add the property set to the XML tree."""
        # check if at least one parameter is specified (case-insensitive)
        # and build a map from material to Matml properties
        available_mat_properties = [model.name.lower() for model in material.models]
        property_set_parameters = {item: item for item in parameter_map["properties"]}
        for key, mapped_properties in parameter_map["mappings"].items():
            property_set_parameters.update({item: key for item in mapped_properties})

        # build final map with property name in MaterialManager to Matml
        parameters = {
            param: key
            for param, key in property_set_parameters.items()
            if param.lower() in available_mat_properties
        }

        if len(parameters) > 0:
            # get property id from metadata or add it if it does not exist yet
            if property_set_name in self._metadata_property_sets.keys():
                property_id = self._metadata_property_sets[property_set_name]
            else:
                index = len(self._metadata_property_sets) + 1
                property_id = f"pr{index}"
                self._metadata_property_sets[property_set_name] = property_id

            property_data_element = ET.SubElement(
                bulkdata_element, "PropertyData", {"property": property_id}
            )
            data_element = ET.SubElement(property_data_element, "Data", {"format": "string"})
            data_element.text = "-"
            if behavior:
                behavior_element = ET.SubElement(
                    property_data_element, "Qualifier", {"name": "Behavior"}
                )
                behavior_element.text = behavior

            if property_set_name == "Coefficient of Thermal Expansion":
                qualifier_element = ET.SubElement(
                    property_data_element, "Qualifier", {"name": "Definition"}
                )
                qualifier_element.text = "Secant"
            self._add_parameters(property_data_element, material, parameters)

    def _add_materials(self, materials_element: ET.Element):
        """Add the material data to the XML tree."""
        for material in self._materials:
            mat_element = ET.SubElement(materials_element, "Material")
            bulkdata_element = ET.SubElement(mat_element, BULKDATA_KEY)
            name_element = ET.SubElement(bulkdata_element, "Name")
            name_element.text = material.name

            for property_set_name, parameters in MATML_PROPERTY_MAP.items():
                # property sets are exported as orthotropic if it can have an isotropic or
                # orthotropic representation,
                if len(property_set_name.split("::")) == 2:
                    behavior = property_set_name.split("::")[1]
                else:
                    behavior = ""
                if behavior != "Isotropic":
                    self._add_property_set(
                        bulkdata_element,
                        material,
                        property_set_name.split("::")[0],
                        parameters,
                        behavior,
                    )

    def _add_metadata(self, metadata_element: ET.Element):
        # add the metadata to the XML tree
        for key, value in self._metadata_property_sets.items():
            prop_element = ET.SubElement(metadata_element, "PropertyDetails", {"id": value})
            ET.SubElement(prop_element, UNITLESS_KEY)
            name_element = ET.SubElement(prop_element, "Name")
            name_element.text = key

        for key, value in self._metadata_parameters.items():
            prop_element = ET.SubElement(metadata_element, "ParameterDetails", {"id": value})
            ET.SubElement(prop_element, UNITLESS_KEY)
            name_element = ET.SubElement(prop_element, "Name")
            name_element.text = key

    def _add_transfer_ids(self, root: ET.Element) -> None:
        # add the WB transfer IDs to the XML tree
        wb_transfer_element = ET.SubElement(root, WBTRANSFER_KEY)
        materials_element = ET.SubElement(wb_transfer_element, MATERIALS_ELEMENT_KEY)
        for mat in self._materials:
            mat_element = ET.SubElement(materials_element, "Material")
            name_element = ET.SubElement(mat_element, "Name")
            name_element.text = mat.name
            transfer_element = ET.SubElement(mat_element, "DataTransferID")
            transfer_element.text = mat.uuid

    def _to_etree(self) -> ET.ElementTree:
        root = ET.Element(ROOT_ELEMENT)
        tree = ET.ElementTree(root)

        root.attrib["version"] = VERSION
        root.attrib["versiondate"] = VERSION_DATE
        notes_element = ET.SubElement(root, "Notes")
        notes_element.text = "Engineering data xml file generated by pyMaterials."

        materials_element = ET.SubElement(root, MATERIALS_ELEMENT_KEY)
        matml_doc_element = ET.SubElement(materials_element, MATML_DOC_KEY)

        self._add_materials(matml_doc_element)

        # add metadata to the XML tree
        metadata_element = ET.SubElement(matml_doc_element, METADATA_KEY)
        self._add_metadata(metadata_element)

        # add transfer id to the XML tree
        self._add_transfer_ids(root)
        return tree

    def _indent(self, tree) -> None:
        if hasattr(ET, "indent"):
            ET.indent(tree)
        else:
            print(f"ElementTree does not have `indent`. Python 3.9+ required!")

    def _xml_write_kwargs(self, **kwargs) -> dict:
        if sys.version_info.minor == 7:
            return {}
        return {"xml_declaration": kwargs.get("xml_declaration", False)}

    def write(self, buffer: BinaryIO, **kwargs) -> None:
        """
        Write a Matml (engineering data xml format) representation of materials to buffer.

        Parameters
        ----------
        buffer:
            buffer to write to
        **kwargs : bool, optional
            Optional keyword arguments.
            indent : bool, optional
                Whether to add an indent to format the XML output(python 3.9+).
                Defaults to ``false``.
            xml_declaration: bool, optional
                Whether to add the XML declaration to the output
        """
        tree = self._to_etree()

        if kwargs.get("indent", False):
            self._indent(tree)
        buffer.write(ET.tostring(tree.getroot(), **self._xml_write_kwargs(**kwargs)))

    def export(self, path: _PATH_TYPE, **kwargs) -> None:
        """
        Write a Matml (engineering data xml format) representation of materials to file.

        Parameters
        ----------
        path:
            File path
        **kwargs : bool, optional
            Optional keyword arguments.
            indent : bool, optional
                Whether to add an indent to format the XML output(python 3.9+).
                Defaults to ``false``.
            xml_declaration: bool, optional
                Whether to add the XML declaration to the output
        """
        tree = self._to_etree()

        print(f"write xml to {path}")
        if kwargs.get("indent", False):
            self._indent(tree)
        tree.write(path, **self._xml_write_kwargs(**kwargs))
